@@ -1,18 +1,12 @@
-/* This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-
 use crate::{
-    device::{
-        descriptor::{DescriptorSet, DescriptorTotalCount},
-        DeviceError, SHADER_STAGE_COUNT,
-    },
+    device::{DeviceError, MissingDownlevelFlags, MissingFeatures, SHADER_STAGE_COUNT},
+    error::{ErrorFormatter, PrettyError},
     hub::Resource,
     id::{BindGroupLayoutId, BufferId, DeviceId, SamplerId, TextureViewId, Valid},
     memory_init_tracker::MemoryInitTrackerAction,
     track::{TrackerSet, UsageConflict, DUMMY_SELECTOR},
     validation::{MissingBufferUsageError, MissingTextureUsageError},
-    FastHashMap, Label, LifeGuard, MultiRefCount, Stored, MAX_BIND_GROUPS,
+    FastHashMap, Label, LifeGuard, MultiRefCount, Stored,
 };
 
 use arrayvec::ArrayVec;
@@ -29,16 +23,28 @@ use std::{
 
 use thiserror::Error;
 
+#[derive(Clone, Debug, Error)]
+pub enum BindGroupLayoutEntryError {
+    #[error("arrays of bindings unsupported for this type of binding")]
+    ArrayUnsupported,
+    #[error(transparent)]
+    MissingFeatures(#[from] MissingFeatures),
+    #[error(transparent)]
+    MissingDownlevelFlags(#[from] MissingDownlevelFlags),
+}
+
 #[derive(Clone, Debug, Error)]
 pub enum CreateBindGroupLayoutError {
     #[error(transparent)]
     Device(#[from] DeviceError),
-    #[error("arrays of bindings unsupported for this type of binding")]
-    ArrayUnsupported,
     #[error("conflicting binding at index {0}")]
     ConflictBinding(u32),
-    #[error("required device feature is missing: {0:?}")]
-    MissingFeature(wgt::Features),
+    #[error("binding {binding} entry is invalid")]
+    Entry {
+        binding: u32,
+        #[source]
+        error: BindGroupLayoutEntryError,
+    },
     #[error(transparent)]
     TooManyBindings(BindingTypeMaxCountError),
 }
@@ -83,12 +89,8 @@ pub enum CreateBindGroupError {
     MissingBufferUsage(#[from] MissingBufferUsageError),
     #[error(transparent)]
     MissingTextureUsage(#[from] MissingTextureUsageError),
-    #[error("required device features not enabled: {0:?}")]
-    MissingFeatures(wgt::Features),
     #[error("binding declared as a single item, but bind group is using it as an array")]
     SingleBindingExpected,
-    #[error("unable to create a bind group with a swap chain image")]
-    SwapChainImage,
     #[error("buffer offset {0} does not respect `BIND_BUFFER_ALIGNMENT`")]
     UnalignedBufferOffset(wgt::BufferAddress),
     #[error(
@@ -146,16 +148,43 @@ pub enum CreateBindGroupError {
     },
     #[error("bound texture views can not have both depth and stencil aspects enabled")]
     DepthStencilAspect,
-    #[error("the adapter does not support simultaneous read + write storage texture access for the format {0:?}")]
-    StorageReadWriteNotSupported(wgt::TextureFormat),
+    #[error("the adapter does not support read access for storages texture of format {0:?}")]
+    StorageReadNotSupported(wgt::TextureFormat),
     #[error(transparent)]
     ResourceUsageConflict(#[from] UsageConflict),
 }
 
+impl PrettyError for CreateBindGroupError {
+    fn fmt_pretty(&self, fmt: &mut ErrorFormatter) {
+        fmt.error(self);
+        match *self {
+            Self::BindingZeroSize(id) => {
+                fmt.buffer_label(&id);
+            }
+            Self::BindingRangeTooLarge { buffer, .. } => {
+                fmt.buffer_label(&buffer);
+            }
+            Self::BindingSizeTooSmall { buffer, .. } => {
+                fmt.buffer_label(&buffer);
+            }
+            Self::InvalidBuffer(id) => {
+                fmt.buffer_label(&id);
+            }
+            Self::InvalidTextureView(id) => {
+                fmt.texture_view_label(&id);
+            }
+            Self::InvalidSampler(id) => {
+                fmt.sampler_label(&id);
+            }
+            _ => {}
+        };
+    }
+}
+
 #[derive(Clone, Debug, Error)]
 pub enum BindingZone {
     #[error("stage {0:?}")]
-    Stage(wgt::ShaderStage),
+    Stage(wgt::ShaderStages),
     #[error("whole pipeline")]
     Pipeline,
 }
@@ -188,29 +217,29 @@ pub(crate) struct PerStageBindingTypeCounter {
 }
 
 impl PerStageBindingTypeCounter {
-    pub(crate) fn add(&mut self, stage: wgt::ShaderStage, count: u32) {
-        if stage.contains(wgt::ShaderStage::VERTEX) {
+    pub(crate) fn add(&mut self, stage: wgt::ShaderStages, count: u32) {
+        if stage.contains(wgt::ShaderStages::VERTEX) {
             self.vertex += count;
         }
-        if stage.contains(wgt::ShaderStage::FRAGMENT) {
+        if stage.contains(wgt::ShaderStages::FRAGMENT) {
             self.fragment += count;
         }
-        if stage.contains(wgt::ShaderStage::COMPUTE) {
+        if stage.contains(wgt::ShaderStages::COMPUTE) {
             self.compute += count;
         }
     }
 
     pub(crate) fn max(&self) -> (BindingZone, u32) {
         let max_value = self.vertex.max(self.fragment.max(self.compute));
-        let mut stage = wgt::ShaderStage::NONE;
+        let mut stage = wgt::ShaderStages::NONE;
         if max_value == self.vertex {
-            stage |= wgt::ShaderStage::VERTEX
+            stage |= wgt::ShaderStages::VERTEX
         }
         if max_value == self.fragment {
-            stage |= wgt::ShaderStage::FRAGMENT
+            stage |= wgt::ShaderStages::FRAGMENT
         }
         if max_value == self.compute {
-            stage |= wgt::ShaderStage::COMPUTE
+            stage |= wgt::ShaderStages::COMPUTE
         }
         (BindingZone::Stage(stage), max_value)
     }
@@ -381,19 +410,18 @@ pub struct BindGroupLayoutDescriptor<'a> {
 pub(crate) type BindEntryMap = FastHashMap<u32, wgt::BindGroupLayoutEntry>;
 
 #[derive(Debug)]
-pub struct BindGroupLayout<B: hal::Backend> {
-    pub(crate) raw: B::DescriptorSetLayout,
+pub struct BindGroupLayout<A: hal::Api> {
+    pub(crate) raw: A::BindGroupLayout,
     pub(crate) device_id: Stored<DeviceId>,
     pub(crate) multi_ref_count: MultiRefCount,
     pub(crate) entries: BindEntryMap,
-    pub(crate) desc_count: DescriptorTotalCount,
     pub(crate) dynamic_count: usize,
     pub(crate) count_validator: BindingTypeMaxCountValidator,
     #[cfg(debug_assertions)]
     pub(crate) label: String,
 }
 
-impl<B: hal::Backend> Resource for BindGroupLayout<B> {
+impl<A: hal::Api> Resource for BindGroupLayout<A> {
     const TYPE: &'static str = "BindGroupLayout";
 
     fn life_guard(&self) -> &LifeGuard {
@@ -419,13 +447,13 @@ pub enum CreatePipelineLayoutError {
         wgt::PUSH_CONSTANT_ALIGNMENT
     )]
     MisalignedPushConstantRange { index: usize, bound: u32 },
-    #[error("device does not have required feature: {0:?}")]
-    MissingFeature(wgt::Features),
+    #[error(transparent)]
+    MissingFeatures(#[from] MissingFeatures),
     #[error("push constant range (index {index}) provides for stage(s) {provided:?} but there exists another range that provides stage(s) {intersected:?}. Each stage may only be provided by one range")]
     MoreThanOnePushConstantRangePerStage {
         index: usize,
-        provided: wgt::ShaderStage,
-        intersected: wgt::ShaderStage,
+        provided: wgt::ShaderStages,
+        intersected: wgt::ShaderStages,
     },
     #[error("push constant at index {index} has range {}..{} which exceeds device push constant size limit 0..{max}", range.start, range.end)]
     PushConstantRangeTooLarge {
@@ -439,6 +467,15 @@ pub enum CreatePipelineLayoutError {
     TooManyGroups { actual: usize, max: usize },
 }
 
+impl PrettyError for CreatePipelineLayoutError {
+    fn fmt_pretty(&self, fmt: &mut ErrorFormatter) {
+        fmt.error(self);
+        if let Self::InvalidBindGroupLayout(id) = *self {
+            fmt.bind_group_layout_label(&id);
+        };
+    }
+}
+
 #[derive(Clone, Debug, Error)]
 pub enum PushConstantUploadError {
     #[error("provided push constant with indices {offset}..{end_offset} overruns matching push constant range at index {idx}, with stage(s) {:?} and indices {:?}", range.stages, range.range)]
@@ -450,20 +487,20 @@ pub enum PushConstantUploadError {
     },
     #[error("provided push constant is for stage(s) {actual:?}, stage with a partial match found at index {idx} with stage(s) {matched:?}, however push constants must be complete matches")]
     PartialRangeMatch {
-        actual: wgt::ShaderStage,
+        actual: wgt::ShaderStages,
         idx: usize,
-        matched: wgt::ShaderStage,
+        matched: wgt::ShaderStages,
     },
     #[error("provided push constant is for stage(s) {actual:?}, but intersects a push constant range (at index {idx}) with stage(s) {missing:?}. Push constants must provide the stages for all ranges they intersect")]
     MissingStages {
-        actual: wgt::ShaderStage,
+        actual: wgt::ShaderStages,
         idx: usize,
-        missing: wgt::ShaderStage,
+        missing: wgt::ShaderStages,
     },
     #[error("provided push constant is for stage(s) {actual:?}, however the pipeline layout has no push constant range for the stage(s) {unmatched:?}")]
     UnmatchedStages {
-        actual: wgt::ShaderStage,
-        unmatched: wgt::ShaderStage,
+        actual: wgt::ShaderStages,
+        unmatched: wgt::ShaderStages,
     },
     #[error("provided push constant offset {0} does not respect `PUSH_CONSTANT_ALIGNMENT`")]
     Unaligned(u32),
@@ -490,19 +527,19 @@ pub struct PipelineLayoutDescriptor<'a> {
 }
 
 #[derive(Debug)]
-pub struct PipelineLayout<B: hal::Backend> {
-    pub(crate) raw: B::PipelineLayout,
+pub struct PipelineLayout<A: hal::Api> {
+    pub(crate) raw: A::PipelineLayout,
     pub(crate) device_id: Stored<DeviceId>,
     pub(crate) life_guard: LifeGuard,
-    pub(crate) bind_group_layout_ids: ArrayVec<[Valid<BindGroupLayoutId>; MAX_BIND_GROUPS]>,
-    pub(crate) push_constant_ranges: ArrayVec<[wgt::PushConstantRange; SHADER_STAGE_COUNT]>,
+    pub(crate) bind_group_layout_ids: ArrayVec<Valid<BindGroupLayoutId>, { hal::MAX_BIND_GROUPS }>,
+    pub(crate) push_constant_ranges: ArrayVec<wgt::PushConstantRange, { SHADER_STAGE_COUNT }>,
 }
 
-impl<B: hal::Backend> PipelineLayout<B> {
+impl<A: hal::Api> PipelineLayout<A> {
     /// Validate push constants match up with expected ranges.
     pub(crate) fn validate_push_constant_ranges(
         &self,
-        stages: wgt::ShaderStage,
+        stages: wgt::ShaderStages,
         offset: u32,
         end_offset: u32,
     ) -> Result<(), PushConstantUploadError> {
@@ -533,7 +570,7 @@ impl<B: hal::Backend> PipelineLayout<B> {
         //   when we check for 1, we can simply check that our entire updated range
         //   is within a push constant range. i.e. our range for a specific stage cannot
         //   intersect more than one push constant range.
-        let mut used_stages = wgt::ShaderStage::NONE;
+        let mut used_stages = wgt::ShaderStages::NONE;
         for (idx, range) in self.push_constant_ranges.iter().enumerate() {
             // contains not intersects due to 2
             if stages.contains(range.stages) {
@@ -578,7 +615,7 @@ impl<B: hal::Backend> PipelineLayout<B> {
     }
 }
 
-impl<B: hal::Backend> Resource for PipelineLayout<B> {
+impl<A: hal::Api> Resource for PipelineLayout<A> {
     const TYPE: &'static str = "PipelineLayout";
 
     fn life_guard(&self) -> &LifeGuard {
@@ -628,8 +665,8 @@ pub struct BindGroupDynamicBindingData {
 }
 
 #[derive(Debug)]
-pub struct BindGroup<B: hal::Backend> {
-    pub(crate) raw: DescriptorSet<B>,
+pub struct BindGroup<A: hal::Api> {
+    pub(crate) raw: A::BindGroup,
     pub(crate) device_id: Stored<DeviceId>,
     pub(crate) layout_id: Valid<BindGroupLayoutId>,
     pub(crate) life_guard: LifeGuard,
@@ -638,7 +675,7 @@ pub struct BindGroup<B: hal::Backend> {
     pub(crate) dynamic_binding_info: Vec<BindGroupDynamicBindingData>,
 }
 
-impl<B: hal::Backend> BindGroup<B> {
+impl<A: hal::Api> BindGroup<A> {
     pub(crate) fn validate_dynamic_bindings(
         &self,
         offsets: &[wgt::DynamicOffset],
@@ -673,13 +710,13 @@ impl<B: hal::Backend> BindGroup<B> {
     }
 }
 
-impl<B: hal::Backend> Borrow<()> for BindGroup<B> {
+impl<A: hal::Api> Borrow<()> for BindGroup<A> {
     fn borrow(&self) -> &() {
         &DUMMY_SELECTOR
     }
 }
 
-impl<B: hal::Backend> Resource for BindGroup<B> {
+impl<A: hal::Api> Resource for BindGroup<A> {
     const TYPE: &'static str = "BindGroup";
 
     fn life_guard(&self) -> &LifeGuard {
