@@ -1,7 +1,3 @@
-/* This Source Code Form is subject to the terms of the Mozilla Public
- * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
-
 use crate::{binding_model::BindEntryMap, FastHashMap, FastHashSet};
 use naga::valid::GlobalUse;
 use std::{collections::hash_map::Entry, fmt};
@@ -111,6 +107,7 @@ struct EntryPoint {
 
 #[derive(Debug)]
 pub struct Interface {
+    features: wgt::Features,
     resources: naga::Arena<Resource>,
     entry_points: FastHashMap<(naga::ShaderStage, String), EntryPoint>,
 }
@@ -118,15 +115,15 @@ pub struct Interface {
 #[derive(Clone, Debug, Error)]
 #[error("buffer usage is {actual:?} which does not contain required usage {expected:?}")]
 pub struct MissingBufferUsageError {
-    pub(crate) actual: wgt::BufferUsage,
-    pub(crate) expected: wgt::BufferUsage,
+    pub(crate) actual: wgt::BufferUsages,
+    pub(crate) expected: wgt::BufferUsages,
 }
 
 /// Checks that the given buffer usage contains the required buffer usage,
 /// returns an error otherwise.
 pub fn check_buffer_usage(
-    actual: wgt::BufferUsage,
-    expected: wgt::BufferUsage,
+    actual: wgt::BufferUsages,
+    expected: wgt::BufferUsages,
 ) -> Result<(), MissingBufferUsageError> {
     if !actual.contains(expected) {
         Err(MissingBufferUsageError { actual, expected })
@@ -138,15 +135,15 @@ pub fn check_buffer_usage(
 #[derive(Clone, Debug, Error)]
 #[error("texture usage is {actual:?} which does not contain required usage {expected:?}")]
 pub struct MissingTextureUsageError {
-    pub(crate) actual: wgt::TextureUsage,
-    pub(crate) expected: wgt::TextureUsage,
+    pub(crate) actual: wgt::TextureUsages,
+    pub(crate) expected: wgt::TextureUsages,
 }
 
 /// Checks that the given texture usage contains the required texture usage,
 /// returns an error otherwise.
 pub fn check_texture_usage(
-    actual: wgt::TextureUsage,
-    expected: wgt::TextureUsage,
+    actual: wgt::TextureUsages,
+    expected: wgt::TextureUsages,
 ) -> Result<(), MissingTextureUsageError> {
     if !actual.contains(expected) {
         Err(MissingTextureUsageError { actual, expected })
@@ -192,6 +189,10 @@ pub enum BindingError {
     InconsistentlyDerivedType,
     #[error("texture format {0:?} is not supported for storage use")]
     BadStorageFormat(wgt::TextureFormat),
+    #[error(
+        "storage texture usage {0:?} doesn't have a matching supported `StorageTextureAccess`"
+    )]
+    UnsupportedTextureStorageAccess(GlobalUse),
 }
 
 #[derive(Clone, Debug, Error)]
@@ -324,7 +325,14 @@ impl Resource {
                             wgt::BufferBindingType::Storage { read_only } => {
                                 let mut global_use = GlobalUse::READ | GlobalUse::QUERY;
                                 global_use.set(GlobalUse::WRITE, !read_only);
-                                (naga::StorageClass::Storage, global_use)
+                                let mut naga_access = naga::StorageAccess::LOAD;
+                                naga_access.set(naga::StorageAccess::STORE, !read_only);
+                                (
+                                    naga::StorageClass::Storage {
+                                        access: naga_access,
+                                    },
+                                    global_use,
+                                )
                             }
                         };
                         if self.class != class {
@@ -420,7 +428,7 @@ impl Resource {
                                 kind: naga::ScalarKind::Uint,
                                 multi,
                             },
-                            wgt::TextureSampleType::Depth => naga::ImageClass::Depth,
+                            wgt::TextureSampleType::Depth => naga::ImageClass::Depth { multi },
                         };
                         (class, GlobalUse::READ | GlobalUse::QUERY)
                     }
@@ -431,16 +439,26 @@ impl Resource {
                     } => {
                         let naga_format = map_storage_format_to_naga(format)
                             .ok_or(BindingError::BadStorageFormat(format))?;
-                        let usage = match access {
-                            wgt::StorageTextureAccess::ReadOnly => {
-                                GlobalUse::READ | GlobalUse::QUERY
+                        let (naga_access, usage) = match access {
+                            wgt::StorageTextureAccess::ReadOnly => (
+                                naga::StorageAccess::LOAD,
+                                GlobalUse::READ | GlobalUse::QUERY,
+                            ),
+                            wgt::StorageTextureAccess::WriteOnly => (
+                                naga::StorageAccess::STORE,
+                                GlobalUse::WRITE | GlobalUse::QUERY,
+                            ),
+                            wgt::StorageTextureAccess::ReadWrite => {
+                                (naga::StorageAccess::all(), GlobalUse::all())
                             }
-                            wgt::StorageTextureAccess::WriteOnly => {
-                                GlobalUse::WRITE | GlobalUse::QUERY
-                            }
-                            wgt::StorageTextureAccess::ReadWrite => GlobalUse::all(),
                         };
-                        (naga::ImageClass::Storage(naga_format), usage)
+                        (
+                            naga::ImageClass::Storage {
+                                format: naga_format,
+                                access: naga_access,
+                            },
+                            usage,
+                        )
                     }
                     _ => return Err(BindingError::WrongType),
                 };
@@ -464,12 +482,16 @@ impl Resource {
         }
     }
 
-    fn derive_binding_type(&self, shader_usage: GlobalUse) -> Result<BindingType, BindingError> {
+    fn derive_binding_type(
+        &self,
+        shader_usage: GlobalUse,
+        features: wgt::Features,
+    ) -> Result<BindingType, BindingError> {
         Ok(match self.ty {
             ResourceType::Buffer { size } => BindingType::Buffer {
                 ty: match self.class {
                     naga::StorageClass::Uniform => wgt::BufferBindingType::Uniform,
-                    naga::StorageClass::Storage => wgt::BufferBindingType::Storage {
+                    naga::StorageClass::Storage { .. } => wgt::BufferBindingType::Storage {
                         read_only: !shader_usage.contains(GlobalUse::WRITE),
                     },
                     _ => return Err(BindingError::WrongType),
@@ -507,16 +529,24 @@ impl Resource {
                         view_dimension,
                         multisampled: multi,
                     },
-                    naga::ImageClass::Depth => BindingType::Texture {
+                    naga::ImageClass::Depth { multi } => BindingType::Texture {
                         sample_type: wgt::TextureSampleType::Depth,
                         view_dimension,
-                        multisampled: false,
+                        multisampled: multi,
                     },
-                    naga::ImageClass::Storage(format) => BindingType::StorageTexture {
-                        access: if shader_usage.contains(GlobalUse::WRITE) {
+                    naga::ImageClass::Storage { format, .. } => BindingType::StorageTexture {
+                        access: if shader_usage == GlobalUse::WRITE || shader_usage.is_empty() {
                             wgt::StorageTextureAccess::WriteOnly
-                        } else {
+                        } else if !features
+                            .contains(wgt::Features::TEXTURE_ADAPTER_SPECIFIC_FORMAT_FEATURES)
+                        {
+                            return Err(BindingError::UnsupportedTextureStorageAccess(
+                                shader_usage,
+                            ));
+                        } else if shader_usage == GlobalUse::READ {
                             wgt::StorageTextureAccess::ReadOnly
+                        } else {
+                            wgt::StorageTextureAccess::ReadWrite
                         },
                         view_dimension,
                         format: {
@@ -621,6 +651,7 @@ impl NumericType {
             Tf::Depth32Float | Tf::Depth24Plus | Tf::Depth24PlusStencil8 => {
                 panic!("Unexpected depth format")
             }
+            Tf::Rgb9e5Ufloat => (NumericDimension::Vector(Vs::Tri), Sk::Float),
             Tf::Bc1RgbaUnorm
             | Tf::Bc1RgbaUnormSrgb
             | Tf::Bc2RgbaUnorm
@@ -631,8 +662,6 @@ impl NumericType {
             | Tf::Bc7RgbaUnormSrgb
             | Tf::Etc2RgbA1Unorm
             | Tf::Etc2RgbA1UnormSrgb
-            | Tf::Etc2RgbA8Unorm
-            | Tf::Etc2RgbA8UnormSrgb
             | Tf::Astc4x4RgbaUnorm
             | Tf::Astc4x4RgbaUnormSrgb
             | Tf::Astc5x4RgbaUnorm
@@ -664,7 +693,7 @@ impl NumericType {
             Tf::Bc4RUnorm | Tf::Bc4RSnorm | Tf::EacRUnorm | Tf::EacRSnorm => {
                 (NumericDimension::Scalar, Sk::Float)
             }
-            Tf::Bc5RgUnorm | Tf::Bc5RgSnorm | Tf::EtcRgUnorm | Tf::EtcRgSnorm => {
+            Tf::Bc5RgUnorm | Tf::Bc5RgSnorm | Tf::EacRgUnorm | Tf::EacRgSnorm => {
                 (NumericDimension::Vector(Vs::Bi), Sk::Float)
             }
             Tf::Bc6hRgbUfloat | Tf::Bc6hRgbSfloat | Tf::Etc2RgbUnorm | Tf::Etc2RgbUnormSrgb => {
@@ -789,7 +818,11 @@ impl Interface {
         list.push(varying);
     }
 
-    pub fn new(module: &naga::Module, info: &naga::valid::ModuleInfo) -> Self {
+    pub fn new(
+        module: &naga::Module,
+        info: &naga::valid::ModuleInfo,
+        features: wgt::Features,
+    ) -> Self {
         let mut resources = naga::Arena::new();
         let mut resource_mapping = FastHashMap::default();
         for (var_handle, var) in module.global_variables.iter() {
@@ -820,12 +853,15 @@ impl Interface {
                     continue;
                 }
             };
-            let handle = resources.append(Resource {
-                name: var.name.clone(),
-                bind,
-                ty,
-                class: var.class,
-            });
+            let handle = resources.append(
+                Resource {
+                    name: var.name.clone(),
+                    bind,
+                    ty,
+                    class: var.class,
+                },
+                Default::default(),
+            );
             resource_mapping.insert(var_handle, handle);
         }
 
@@ -860,6 +896,7 @@ impl Interface {
         }
 
         Interface {
+            features,
             resources,
             entry_points,
         }
@@ -870,15 +907,15 @@ impl Interface {
         given_layouts: Option<&[&BindEntryMap]>,
         derived_layouts: &mut [BindEntryMap],
         entry_point_name: &str,
-        stage_bit: wgt::ShaderStage,
+        stage_bit: wgt::ShaderStages,
         inputs: StageIo,
     ) -> Result<StageIo, StageError> {
         // Since a shader module can have multiple entry points with the same name,
         // we need to look for one with the right execution model.
         let shader_stage = match stage_bit {
-            wgt::ShaderStage::VERTEX => naga::ShaderStage::Vertex,
-            wgt::ShaderStage::FRAGMENT => naga::ShaderStage::Fragment,
-            wgt::ShaderStage::COMPUTE => naga::ShaderStage::Compute,
+            wgt::ShaderStages::VERTEX => naga::ShaderStage::Vertex,
+            wgt::ShaderStages::FRAGMENT => naga::ShaderStage::Fragment,
+            wgt::ShaderStages::COMPUTE => naga::ShaderStage::Compute,
             _ => unreachable!(),
         };
         let pair = (shader_stage, entry_point_name.to_string());
@@ -907,7 +944,7 @@ impl Interface {
                     .get_mut(res.bind.group as usize)
                     .ok_or(BindingError::Missing)
                     .and_then(|set| {
-                        let ty = res.derive_binding_type(usage)?;
+                        let ty = res.derive_binding_type(usage, self.features)?;
                         match set.entry(res.bind.binding) {
                             Entry::Occupied(e) if e.get().ty != ty => {
                                 return Err(BindingError::InconsistentlyDerivedType)
